from drf_stripe.models import Product
from drf_stripe.stripe_api.products import create_update_product_features
from drf_stripe.stripe_models.product import StripeProductEventData


def _handle_product_event_data(data: StripeProductEventData):
    product_id = data.object.id
    active = data.object.active
    description = data.object.description
    name = data.object.name

    product, created = Product.objects.update_or_create(product_id=product_id, defaults={
        "active": active,
        "description": description,
        "name": name
    })

    create_update_product_features(data.object)



from datetime import datetime
from enum import Enum
from typing import List, Dict, Optional

from pydantic import BaseModel

from .price import StripePrice


class StripeSubscriptionStatus(str, Enum):
    """See: https://stripe.com/docs/api/subscriptions/object#subscription_object-status"""
    ACTIVE = 'active'
    PAST_DUE = 'past_due'
    UNPAID = 'unpaid'
    CANCELED = 'canceled'
    INCOMPLETE = 'incomplete'
    INCOMPLETE_EXPIRED = 'incomplete_expired'
    TRIALING = 'trialing'
    ENDED = 'ended'


ACCESS_GRANTING_STATUSES = (
    StripeSubscriptionStatus.ACTIVE,
    StripeSubscriptionStatus.PAST_DUE,
    StripeSubscriptionStatus.TRIALING
)


class StripeSubscriptionItemsDataItem(BaseModel):
    """Based on https://stripe.com/docs/api/subscriptions/object#subscription_object-items-data"""
    id: str
    billing_thresholds: Dict = None
    created: datetime
    metadata: Dict
    price: StripePrice
    quantity: int
    subscription: str
    tax_rates: List


class StripeSubscriptionItems(BaseModel):
    """Based on https://stripe.com/docs/api/subscriptions/object#subscription_object-items"""
    data: List[StripeSubscriptionItemsDataItem]
    has_more: bool = None
    url: str = None


class StripeSubscription(BaseModel):
    """Based on https://stripe.com/docs/api/subscriptions/object"""
    id: Optional[str]
    cancel_at_period_end: Optional[bool]
    cancel_at: Optional[datetime] = None
    ended_at: Optional[datetime] = None
    trial_end: Optional[datetime] = None
    trial_start: Optional[datetime] = None
    current_period_end: Optional[datetime]
    current_period_start: Optional[datetime]
    customer: Optional[str]
    default_payment_method: str = None
    items: Optional[StripeSubscriptionItems]
    latest_invoice: Optional[str]
    metadata: Optional[Dict]
    pending_setup_intent: str = None
    pending_update: str = None
    status: Optional[StripeSubscriptionStatus]


class StripeSubscriptions(BaseModel):
    """Based on https://stripe.com/docs/api/subscriptions/list"""
    data: List[StripeSubscription]
    has_more: bool = None
    url: str = None


class StripeSubscriptionEventData(BaseModel):
    """Based on https://stripe.com/docs/api/events/object#event_object-data"""
    object: StripeSubscription
    previous_attributes: Optional[StripeSubscription]



from .api import stripe_api as stripe
from .customers import get_or_create_stripe_user
from ..settings import drf_stripe_settings


def stripe_api_create_billing_portal_session(user_id):
    """
    Creates a Stripe Customer Portal Session.

    :param str user_id: Django User id
    """
    stripe_user = get_or_create_stripe_user(user_id=user_id)

    session = stripe.billing_portal.Session.create(
        customer=stripe_user.customer_id,
        return_url=f"{drf_stripe_settings.FRONT_END_BASE_URL}/manage-subscription/"
    )

    return session



from itertools import chain
from operator import attrgetter
from typing import Literal, List

from django.db.models import Q
from django.db.models import QuerySet
from django.db.transaction import atomic

from drf_stripe.stripe_api.api import stripe_api as stripe
from .customers import get_or_create_stripe_user
from ..models import Subscription, Price, SubscriptionItem
from ..stripe_models.subscription import ACCESS_GRANTING_STATUSES, StripeSubscriptions

"""
status argument, see https://stripe.com/docs/api/subscriptions/list?lang=python#list_subscriptions-status
"""
STATUS_ARG = Literal[
    "active",
    "past_due",
    "unpaid",
    "canceled",
    "incomplete",
    "incomplete_expired",
    "trialing",
    "all",
    "ended"
]


@atomic
def stripe_api_update_subscriptions(status: STATUS_ARG = None, limit: int = 100, starting_after: str = None,
                                    test_data=None):
    """
    Retrieve all subscriptions. Updates database.

    :param STATUS_ARG status: subscription status to retrieve.
    :param int limit: number of instances to retrieve( between 0 and 100).
    :param str starting_after: subscription id to start retrieving.
    :param test_data: response data from Stripe API stripe.Subscription.list, used for testing
    """

    if limit < 0 or limit > 100:
        raise ValueError("Argument limit should be a positive integer no greater than 100.")

    if test_data is None:
        subscriptions_response = stripe.Subscription.list(status=status, limit=limit, starting_after=starting_after)
    else:
        subscriptions_response = test_data

    stripe_subscriptions = StripeSubscriptions(**subscriptions_response).data

    creation_count = 0

    for subscription in stripe_subscriptions:
        stripe_user = get_or_create_stripe_user(customer_id=subscription.customer)

        _, created = Subscription.objects.update_or_create(
            subscription_id=subscription.id,
            defaults={
                "stripe_user": stripe_user,
                "period_start": subscription.current_period_start,
                "period_end": subscription.current_period_end,
                "cancel_at": subscription.cancel_at,
                "cancel_at_period_end": subscription.cancel_at_period_end,
                "ended_at": subscription.ended_at,
                "status": subscription.status,
                "trial_end": subscription.trial_end,
                "trial_start": subscription.trial_start
            }
        )
        print(f"Updated subscription {subscription.id}")
        _update_subscription_items(subscription.id, subscription.items.data)
        if created is True:
            creation_count += 1

    print(f"Created {creation_count} new Subscriptions.")


def _update_subscription_items(subscription_id, items_data):
    SubscriptionItem.objects.filter(subscription=subscription_id).delete()
    for item in items_data:
        _, created = SubscriptionItem.objects.update_or_create(
            sub_item_id=item.id,
            defaults={
                "subscription_id": subscription_id,
                "price_id": item.price.id,
                "quantity": item.quantity
            }
        )
        print(f"Updated sub item {item.id}")


# def _stripe_api_update_subscription_items(subscription_id, limit=100, ending_before=None, test_data=None):
#     """
#     param: str subscription_id: subscription id for which to retrieve subscription items
#     :param int limit: number of instances to retrieve( between 0 and 100).
#     :param str ending_before: subscription item id to retrieve before.
#     """
#     if limit < 0 or limit > 100:
#         raise ValueError("Argument limit should be a positive integer no greater than 100.")
#
#     if test_data is None:
#         items_response = stripe.SubscriptionItem.list(subscription=subscription_id,
#                                                       limit=limit,
#                                                       ending_before=ending_before)
#     else:
#         items_response = test_data
#
#     sub_items = StripeSubscriptionItems(**items_response).data
#
#     SubscriptionItem.objects.filter(subscription=subscription_id).delete()
#     for item in sub_items:
#         SubscriptionItem.objects.update_or_create(
#             sub_item_id=item.id,
#             defaults={
#                 "subscription_id": subscription_id,
#                 "price_id": item.price.id,
#                 "quantity": item.quantity
#             }
#         )


def list_user_subscriptions(user_id, current=True) -> QuerySet[Subscription]:
    """
    Retrieve a set of Subscriptions associated with a given user id.

    :param user_id: Django User id.
    :param bool current: Defaults to True and retrieves only current subscriptions
        (excluding any cancelled, ended, unpaid subscriptions)
    """
    q = Q(stripe_user__user_id=user_id)
    if current is True:
        q &= Q(status__in=ACCESS_GRANTING_STATUSES)

    return Subscription.objects.filter(q)


def list_user_subscription_items(user_id, current=True) -> QuerySet[SubscriptionItem]:
    """
    Retrieve a set of SubscriptionItems associated with user id

    :param user_id: Django User is.
    :param bool current: Defaults to True and retrieves only current subscriptions
        (excluding any cancelled, ended, unpaid subscriptions)
    """
    q = Q(subscription__stripe_user__user_id=user_id)
    if current is True:
        q &= Q(subscription__status__in=ACCESS_GRANTING_STATUSES)

    return SubscriptionItem.objects.filter(q)


def list_user_subscription_products(user_id, current=True):
    """
    Retrieve a set of Product instances associated with a given User instance.

    :param user_id: Django User id.
    :param bool current: Defaults to True and retrieves only products associated with current subscriptions
        (excluding any cancelled, ended, unpaid subscription products)
    """
    subscriptions = list_user_subscriptions(user_id, current=current)
    sub_items = chain.from_iterable(
        sub.items.all() for sub in subscriptions.all().prefetch_related("items__price__product"))
    products = set(item.price.product for item in sub_items)
    return products


def list_subscribable_product_prices_to_user(user_id):
    """
    Retrieve a set of Price instances associated with Products that the User isn't currently subscribed to.

    :param user_id: Django user id.
    """
    current_products = set(map(attrgetter('product_id'), list_user_subscription_products(user_id)))
    prices = Price.objects.filter(
        Q(active=True) &
        Q(product__active=True) &
        ~Q(product__product_id__in=current_products)
    )
    return prices


def list_all_available_product_prices(expand: List = None):
    """Retrieve a set of all Price instances that are available to public."""

    prices = Price.objects.filter(Q(active=True) & Q(product__active=True))

    if expand and "feature" in expand:
        prices = prices.prefetch_related("product__linked_features__feature")

    return prices



from datetime import timedelta, datetime
from functools import reduce
from typing import overload, List, Union
from urllib.parse import urljoin

from django.contrib.auth import get_user_model
from django.utils import timezone

from drf_stripe.stripe_api.api import stripe_api as stripe
from ..settings import drf_stripe_settings


@overload
def stripe_api_create_checkout_session(customer_id: str, price_id: str, trial_end: datetime = None):
    ...


@overload
def stripe_api_create_checkout_session(user_instance, price_id: str, trial_end: datetime = None):
    ...


def stripe_api_create_checkout_session(**kwargs):
    """
    create a Stripe checkout session to start a subscription for user.
    You must provide either customer_id and price_id;
    or user_instance and price_id.
    Optionally provide a trial_end.

    :key user_instance: Django User instance.
    :key customer_id: Stripe customer id.
    :key str price_id: Stripe price id.
    :key int quantity: Defaults to 1.
    :key datetime trial_end: start the subscription with a trial.
    :key list line_items: Used when multiple price + quantity params need to be used. Defaults to None.
        If specified, supersedes price_id and quantity arguments.
    """

    user_instance = kwargs.get("user_instance")
    customer_id = kwargs.get("customer_id")

    if user_instance and isinstance(user_instance, get_user_model()):
        return _stripe_api_create_checkout_session_for_user(**kwargs)
    elif customer_id and isinstance(customer_id, str):
        return _stripe_api_create_checkout_session_for_customer(**kwargs)
    else:
        raise TypeError("Unknown keyword arguments.")


def _stripe_api_create_checkout_session_for_customer(customer_id: str, **kwargs):
    """
    create a Stripe checkout session to start a subscription for user.

    :param customer_id: Stripe customer id.
    :param str price_id: Stripe price id.
    :param int quantity: Defaults to 1.
    :param datetime trial_end: start the subscription with a trial.
    :param list line_items: Used when multiple price + quantity params need to be used. Defaults to None.
        If specified, supersedes price_id and quantity arguments.
    """
    stripe_checkout_params = _make_stripe_checkout_params(customer_id, **kwargs)

    return stripe.checkout.Session.create(**stripe_checkout_params)


def _stripe_api_create_checkout_session_for_user(user_instance, **kwargs):
    """
    create a Stripe checkout session to start a subscription for user.

    :param user_instance: Django User instance.
    :param str price_id: Stripe price id.
    :param bool trial_end: trial_end
    """

    return _stripe_api_create_checkout_session_for_customer(
        customer_id=user_instance.stripe_user.customer_id,
        **kwargs
    )


def _make_stripe_checkout_params(
        customer_id: str, price_id: str = None, quantity: int = 1, line_items: List[dict] = None,
        trial_end: Union[str, datetime, None] = 'auto', discounts: List[dict] = None,
        payment_method_types: List[str] = None, checkout_mode: str = drf_stripe_settings.DEFAULT_CHECKOUT_MODE
):
    if price_id is None and line_items is None:
        raise ValueError("Invalid arguments: must provide either a 'price_id' or 'line_items'.")
    elif price_id is not None and line_items is not None:
        raise ValueError("Invalid arguments: 'price_id' and 'line_items' should be used at the same time.")

    if price_id is not None:
        line_items = [{'price': price_id, 'quantity': quantity}]

    if payment_method_types is None:
        payment_method_types = drf_stripe_settings.DEFAULT_PAYMENT_METHOD_TYPES

    success_url = reduce(urljoin, (drf_stripe_settings.FRONT_END_BASE_URL,
                                   drf_stripe_settings.CHECKOUT_SUCCESS_URL_PATH,
                                   "?session={CHECKOUT_SESSION_ID}"))

    cancel_url = reduce(urljoin, (drf_stripe_settings.FRONT_END_BASE_URL,
                                  drf_stripe_settings.CHECKOUT_CANCEL_URL_PATH))

    params = {
        "customer": customer_id,
        "success_url": success_url,
        "cancel_url": cancel_url,
        "payment_method_types": payment_method_types,
        "mode": checkout_mode,
        "line_items": line_items,
        "subscription_data": {
            "trial_end": _make_trial_end_timestamp(trial_end=trial_end)
        }
    }

    allow_promotion_codes = drf_stripe_settings.ALLOW_PROMOTION_CODES

    if allow_promotion_codes:
        params.update({"allow_promotion_codes": allow_promotion_codes})
    else:
        params.update({"discounts": discounts if discounts else drf_stripe_settings.DEFAULT_DISCOUNTS})

    return params


def _make_trial_end_timestamp(trial_end: Union[str, None, datetime] = 'auto'):
    """
    Returns a new trial_end time to be used for setting up new Stripe Subscription.
    Stripe requires new Subscription trial_end to be at least 48 hours in the future.

    :param trial_end: Explicitly set trial end datetime.
        Defaults to 'auto', which will return a calculated trial_end based on NEW_USER_FREE_TRIAL_DAYS setting.
        If set to less than 48 hours from now, will return the minimum required trial_end acceptable to Stripe API.
        If set to None, no trial, returns None.
    """
    if trial_end is None:
        return

    if trial_end == 'auto':
        if drf_stripe_settings.NEW_USER_FREE_TRIAL_DAYS is None:
            return
        else:
            trial_end = timezone.now() + timezone.timedelta(days=drf_stripe_settings.NEW_USER_FREE_TRIAL_DAYS + 1)

    min_trial_end = timezone.now() + timedelta(hours=49)
    if trial_end < min_trial_end:
        trial_end = min_trial_end

    return int(trial_end.replace(microsecond=0).timestamp())



from django.conf import settings
from django.test.signals import setting_changed

USER_SETTINGS = getattr(settings, "DRF_STRIPE", None)

DEFAULTS = {
    "STRIPE_API_SECRET": "my_stripe_api_key",
    "STRIPE_WEBHOOK_SECRET": "my_stripe_webhook_key",
    "FRONT_END_BASE_URL": "http://localhost:3000",
    "NEW_USER_FREE_TRIAL_DAYS": None,
    "CHECKOUT_SUCCESS_URL_PATH": "payment",
    "CHECKOUT_CANCEL_URL_PATH": "manage-subscription",
    "DEFAULT_PAYMENT_METHOD_TYPES": ["card"],
    "DEFAULT_CHECKOUT_MODE": "subscription",
    "DEFAULT_DISCOUNTS": None,
    "ALLOW_PROMOTION_CODES": True,
}


class DrfStripeSettings:
    def __init__(self, user_settings=None, defaults=None):
        self._user_settings = user_settings or {}
        self.defaults = defaults or DEFAULTS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "DRF_STRIPE", {})
        return self._user_settings

    def __getattr__(self, attr):

        # check the setting is accepted
        if attr not in self.defaults:
            raise AttributeError(f"Invalid DRF_STRIPE setting: {attr}")

        # get from user settings or default value
        try:
            val = self.user_settings[attr]
        except KeyError:
            val = self.defaults[attr]

        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")


drf_stripe_settings = DrfStripeSettings(USER_SETTINGS, DEFAULTS)


def reload_drf_stripe_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "DRF_STRIPE":
        drf_stripe_settings.reload()


setting_changed.connect(reload_drf_stripe_settings)



from drf_stripe.models import Price
from drf_stripe.stripe_api.products import get_freq_from_stripe_price
from drf_stripe.stripe_models.price import StripePriceEventData


def _handle_price_event_data(data: StripePriceEventData):
    price_id = data.object.id
    product_id = data.object.product
    nickname = data.object.nickname
    price = data.object.unit_amount
    active = data.object.active
    freq = get_freq_from_stripe_price(data.object)
    currency = data.object.currency

    price_obj, created = Price.objects.update_or_create(
        price_id=price_id,
        defaults={
            "product_id": product_id,
            "nickname": nickname,
            "price": price,
            "active": active,
            "freq": freq,
            "currency": currency
        }
    )



from django.contrib.auth import get_user_model
from django.db import models

from .stripe_models.subscription import ACCESS_GRANTING_STATUSES


class StripeUser(models.Model):
    """A model linking Django user model with a Stripe User"""
    user = models.OneToOneField(get_user_model(), on_delete=models.CASCADE, related_name='stripe_user',
                                primary_key=True)
    customer_id = models.CharField(max_length=128, null=True)

    @property
    def subscription_items(self):
        """Returns a set of SubscriptionItem instances associated with the StripeUser"""
        return SubscriptionItem.objects.filter(subscription__stripe_user=self)

    @property
    def current_subscription_items(self):
        """Returns a set of SubscriptionItem instances that grants current access."""
        return self.subscription_items.filter(subscription__status__in=ACCESS_GRANTING_STATUSES)

    @property
    def subscribed_products(self):
        """Returns a set of Product instances the StripeUser currently has"""
        return {item.price.product for item in
                self.current_subscription_items.prefetch_related("price", "price__product")}

    @property
    def subscribed_features(self):
        """Returns a set of Feature instances the StripeUser has access to."""
        price_list = self.current_subscription_items.values_list('price', flat=True)
        product_list = Price.objects.filter(pk__in=price_list).values_list("product", flat=True)
        return {item.feature for item in
                ProductFeature.objects.filter(product_id__in=product_list).prefetch_related("feature")}

    class Meta:
        indexes = [
            models.Index(fields=['user', 'customer_id'])
        ]


class Feature(models.Model):
    """
    A model used to keep track of features provided by your application.
    This does not correspond to a Stripe object, but the feature ids should be listed as
     a space delimited strings in Stripe.product.metadata.features
    """
    feature_id = models.CharField(max_length=64, primary_key=True)
    description = models.CharField(max_length=256, null=True, blank=True)


class Product(models.Model):
    """A model representing a Stripe Product"""
    product_id = models.CharField(max_length=256, primary_key=True)
    active = models.BooleanField()
    description = models.CharField(max_length=1024, null=True, blank=True)
    name = models.CharField(max_length=256, null=True, blank=True)


class ProductFeature(models.Model):
    """A model representing association of Product and Feature instances. They have many-to-many relationship."""
    product = models.ForeignKey(Product, on_delete=models.CASCADE, related_name="linked_features")
    feature = models.ForeignKey(Feature, on_delete=models.CASCADE, related_name="linked_products")


class Price(models.Model):
    """A model representing to a Stripe Price object, with enhanced attributes."""
    price_id = models.CharField(max_length=256, primary_key=True)
    product = models.ForeignKey(Product, on_delete=models.CASCADE, related_name="prices")
    nickname = models.CharField(max_length=256, null=True, blank=True)  # displayed name
    price = models.PositiveIntegerField()  # price in cents, corresponding to Stripe unit_amount
    # billing frequency, translated from Stripe price.recurring.interval and price.recurring.interval_count
    freq = models.CharField(max_length=64, null=True, blank=True)
    active = models.BooleanField()
    currency = models.CharField(max_length=3)

    class Meta:
        indexes = [
            models.Index(fields=['active', 'freq'])
        ]


class Subscription(models.Model):
    """
    A model representing Subscription, corresponding to a Stripe Subscription object.
    """
    subscription_id = models.CharField(max_length=256, primary_key=True)
    stripe_user = models.ForeignKey(StripeUser, on_delete=models.CASCADE, related_name="subscriptions")
    period_start = models.DateTimeField(null=True, blank=True)
    period_end = models.DateTimeField(null=True, blank=True)
    cancel_at = models.DateTimeField(null=True, blank=True)
    cancel_at_period_end = models.BooleanField()
    ended_at = models.DateTimeField(null=True, blank=True)
    status = models.CharField(max_length=64)
    trial_end = models.DateTimeField(null=True, blank=True)
    trial_start = models.DateTimeField(null=True, blank=True)

    class Meta:
        indexes = [
            models.Index(fields=['stripe_user', 'status'])
        ]


class SubscriptionItem(models.Model):
    """
    A model representing relation of Price and Subscription, corresponding to Stripe Subscription line item.
    """
    sub_item_id = models.CharField(max_length=256, primary_key=True)
    subscription = models.ForeignKey(Subscription, on_delete=models.CASCADE, related_name="items")
    price = models.ForeignKey(Price, on_delete=models.CASCADE, related_name="+")
    quantity = models.PositiveIntegerField()



from django.core.management.base import BaseCommand

from drf_stripe.stripe_api.customers import stripe_api_update_customers


class Command(BaseCommand):
    help = "Import Stripe Customer objects from Stripe"

    def add_arguments(self, parser):
        parser.add_argument("-l", "--limit", type=int, help="Limit", default=100)
        parser.add_argument("-s", "--starting_after", type=str, help="Starting after customer id", default=None)

    def handle(self, *args, **kwargs):
        stripe_api_update_customers(limit=kwargs.get('limit'), starting_after=kwargs.get('starting_after'))



from django.urls import path

from drf_stripe import views

urlpatterns = [
    path('my-subscription/', views.Subscription.as_view()),
    path('my-subscription-items/', views.SubscriptionItems.as_view()),
    path('subscribable-product/', views.SubscribableProductPrice.as_view()),
    path('checkout/', views.CreateStripeCheckoutSession.as_view()),
    path('webhook/', views.StripeWebhook.as_view()),
    path('customer-portal/', views.StripeCustomerPortal.as_view())
]



from django.db.models import Q
from django.db.transaction import atomic

from drf_stripe.models import Product, Price, Feature, ProductFeature
from .api import stripe_api as stripe
from ..stripe_models.price import StripePrices
from ..stripe_models.product import StripeProducts


@atomic()
def stripe_api_update_products_prices(**kwargs):
    """
    Fetches list of Products and Price from Stripe, updates database.
    :key dict test_products: mock event data for testing
    :key dict test_prices: mock event data for testing
    """
    _stripe_api_fetch_update_products(**kwargs)
    _stripe_api_fetch_update_prices(**kwargs)


def _stripe_api_fetch_update_products(test_products=None, **kwargs):
    """
    Fetch list of Stripe Products and updates database.

    :param dict test_products:  Response from calling Stripe API: stripe.Product.list(). Used for testing.
    """
    if test_products is None:
        products_data = stripe.Product.list(limit=100)
    else:
        products_data = test_products

    products = StripeProducts(**products_data).data

    creation_count = 0
    for product in products:
        product_obj, created = Product.objects.update_or_create(
            product_id=product.id,
            defaults={
                "active": product.active,
                "description": product.description,
                "name": product.name
            }
        )
        create_update_product_features(product)
        if created is True:
            creation_count += 1

    print(f"Created {creation_count} new Products")


def _stripe_api_fetch_update_prices(test_prices=None, **kwargs):
    """
    Fetch list of Stripe Prices and updates database.

    :param dict test_prices: Optional, response from calling Stripe API: stripe.Price.list(). Used for testing.
    """
    if test_prices is None:
        prices_data = stripe.Price.list(limit=100)
    else:
        prices_data = test_prices

    prices = StripePrices(**prices_data).data

    creation_count = 0
    for price in prices:
        price_obj, created = Price.objects.update_or_create(
            price_id=price.id,
            defaults={
                "product_id": price.product,
                "nickname": price.nickname,
                "price": price.unit_amount,
                "freq": get_freq_from_stripe_price(price),
                "active": price.active,
                "currency": price.currency
            }
        )
        if created is True:
            creation_count += 1

    print(f"Created {creation_count} new Prices")


def get_freq_from_stripe_price(price_data):
    """Get 'freq' string from Stripe price data"""
    if price_data.recurring:
        return f"{price_data.recurring.interval}_{price_data.recurring.interval_count}"


@atomic
def create_update_product_features(product_data):
    """
    Create/update Feature instances associated with a Product given product data.
    The features are specified in Stripe Product object metadata.features as space delimited strings.
    See https://stripe.com/docs/api/products/object#product_object-metadata
    """
    if hasattr(product_data, "metadata") and \
            hasattr(product_data.metadata, "features") and \
            product_data.metadata.features:
        features = product_data.metadata.features.split(" ")

        ProductFeature.objects.filter(Q(product_id=product_data.id) & ~Q(feature_id__in=features)).delete()

        for feature_id in features:
            feature_id = feature_id.strip()
            feature, created_new_feature = Feature.objects.get_or_create(
                feature_id=feature_id,
                defaults={"description": feature_id}
            )
            ProductFeature.objects.get_or_create(product_id=product_data.id, feature=feature)

            if created_new_feature:
                print(
                    f"Created new feature_id {feature_id}, please set feature description manually in database.")



from datetime import datetime
from typing import Dict, List, Union, Optional

from pydantic import BaseModel


class PackageDimension(BaseModel):
    height: float = None
    length: float = None
    weight: float = None
    width: float = None


class StripeProductMetadata(BaseModel):
    features: Optional[str] = None


class StripeProduct(BaseModel):
    """A single StripeProduct, see https://stripe.com/docs/api/products/object"""
    id: Optional[str]
    active: Optional[bool]
    description: Optional[str] = None
    metadata: Optional[Union[StripeProductMetadata, Dict]]
    name: Optional[str] = None
    created: Optional[datetime]
    images: Optional[List[str]]
    package_dimensions: Optional[PackageDimension] = None
    shippable: Optional[bool] = None
    statement_descriptor: Optional[str] = None
    tax_code: Optional[Union[str, Dict]] = None
    unit_label: Optional[str] = None
    updated: Optional[datetime] = None
    url: Optional[str] = None


class StripeProducts(BaseModel):
    """List of StripeProducts"""
    url: str
    has_more: bool
    data: List[StripeProduct]


class StripeProductEventData(BaseModel):
    """Based on https://stripe.com/docs/api/products/object"""
    object: StripeProduct
    previous_attributes: Optional[StripeProduct]



from typing import overload

from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db.transaction import atomic

from drf_stripe.models import StripeUser
from drf_stripe.stripe_api.api import stripe_api as stripe
from drf_stripe.stripe_models.customer import StripeCustomers, StripeCustomer


@overload
def get_or_create_stripe_user(user_instance) -> StripeUser:
    ...


@overload
def get_or_create_stripe_user(user_id, user_email) -> StripeUser:
    ...


@overload
def get_or_create_stripe_user(user_id) -> StripeUser:
    ...


@atomic()
def get_or_create_stripe_user(**kwargs) -> StripeUser:
    """
    Get or create a StripeUser given a User instance, or given user id and user email.

    :key user_instance: Django user instance.
    :key str user_id: Django User id.
    :key str user_email: user email address.
    :key str customer_id: Stripe customer id.
    """
    user_instance = kwargs.get("user_instance")
    user_id = kwargs.get("user_id")
    user_email = kwargs.get("user_email")
    customer_id = kwargs.get("customer_id")

    if user_instance and isinstance(user_instance, get_user_model()):
        return _get_or_create_stripe_user_from_user_instance(user_instance)
    elif user_id and user_email and isinstance(user_id, str):
        return _get_or_create_stripe_user_from_user_id_email(user_id, user_email)
    elif user_id is not None:
        return _get_or_create_stripe_user_from_user_id(user_id)
    elif customer_id is not None:
        return _get_or_create_stripe_user_from_customer_id(customer_id)
    else:
        raise TypeError("Unknown keyword arguments!")


def _get_or_create_stripe_user_from_user_instance(user_instance):
    """
    Returns a StripeUser instance given a Django User instance.

    :param user_instance: Django User instance.
    """
    return _get_or_create_stripe_user_from_user_id_email(user_instance.id, user_instance.email)


def _get_or_create_stripe_user_from_user_id(user_id):
    """
    Returns a StripeUser instance given user_id.

    :param str user_id: user id
    """
    user = get_user_model().objects.get(id=user_id)

    return _get_or_create_stripe_user_from_user_id_email(user.id, user.email)


def _get_or_create_stripe_user_from_customer_id(customer_id):
    """
    Returns a StripeUser instance given customer_id

    :param str customer_id: Stripe customer id
    """

    try:
        user = get_user_model().objects.get(stripe_user__customer_id=customer_id)

    except ObjectDoesNotExist:
        customer_response = stripe.Customer.retrieve(customer_id)
        customer = StripeCustomer(**customer_response)
        user, created = get_user_model().objects.get_or_create(
            email=customer.email,
            defaults={"username": customer.email}
        )
        if created:
            print(f"Created new User with customer_id {customer_id}")

    return _get_or_create_stripe_user_from_user_id_email(user.id, user.email)


def _get_or_create_stripe_user_from_user_id_email(user_id, user_email: str):
    """
    Return a StripeUser instance given user_id and user_email.

    :param user_id: user id
    :param str user_email: user email address
    """
    stripe_user, created = StripeUser.objects.get_or_create(user_id=user_id)

    if created:
        customer = _stripe_api_get_or_create_customer_from_email(user_email)
        stripe_user.customer_id = customer.id
        stripe_user.save()

    return stripe_user


def _stripe_api_get_or_create_customer_from_email(user_email: str):
    """
    Get or create a Stripe customer by email address.
    Stripe allows creation of multiple customers with the same email address, therefore it is important that you use
    this method to create or retrieve a Stripe Customer instead of creating one by calling the Stripe API directly.

    :param str user_email: user email address
    """
    customers_response = stripe.Customer.list(email=user_email)
    stripe_customers = StripeCustomers(**customers_response).data

    if len(stripe_customers) > 0:
        customer = stripe_customers.pop()
    else:
        customer = stripe.Customer.create(email=user_email)

    return customer


@atomic
def stripe_api_update_customers(limit=100, starting_after=None, test_data=None):
    """
    Retrieve list of Stripe customer objects, and create Django User and StripeUser instances.

    :param int limit: Limit the number of customers to retrieve
    :param str starting_after: Stripe Customer id to start retrieval
    :param test_data: Stripe.Customer.list API response, used for testing
    """

    if limit < 0 or limit > 100:
        raise ValueError("Argument limit should be a positive integer no greater than 100.")

    if test_data is None:
        customers_response = stripe.Customer.list(limit=limit, starting_after=starting_after)
    else:
        customers_response = test_data

    stripe_customers = StripeCustomers(**customers_response).data

    user_creation_count = 0
    stripe_user_creation_count = 0

    for customer in stripe_customers:
        # Stripe customer can have null as email
        if customer.email is not None:
            user, user_created = get_user_model().objects.get_or_create(
                email=customer.email,
                defaults={"username": customer.email}
            )
            stripe_user, stripe_user_created = StripeUser.objects.get_or_create(user=user,
                                                                                defaults={"customer_id": customer.id})
            print(f"Updated Stripe Customer {customer.id}")

            if user_created is True:
                user_creation_count += 1
            if stripe_user_created is True:
                stripe_user_creation_count += 1

    print(f"{user_creation_count} user(s) created, {stripe_user_creation_count} user(s) linked to Stripe customers.")
